#!/usr/bin/env python
# -*- coding: utf-8 -*-

# GIMP plugin to experiment procedures

# (c) Issabella (tutorial) / Dinasset (filter) / 2021
#
#   History:
#
#   v1.0: 2021-09-30: Initial version
#
#   This program is free software; you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation; either version 2 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program; if not, write to the Free Software
#   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

import sys, os
import random
import time
import datetime
from gimpfu import *
import pygtk
import gtk
from sys import argv, platform
from ctypes import *

# define here some constant value, when required
FixWH=3000 #regular if test=2000

############################################################################
# Service routines
#---------------------------------------------------------------------------

def load_library (library_name):
    if platform == "linux" or platform == "linux2":
        library_name = library_name + '.so.0'
    elif platform == "win32":
        from ctypes.util import find_library
        library_name = find_library (library_name + "-0")
    else:
        raise BaseException ("TODO")
    return CDLL (library_name)
#import sys
#print (sys.version)
gimp = load_library ('libgimp-2.0')
#print (gimp)
#gimp = load_library ('libgimp-3.0')

def gegl_graph(image, drawable, gegl_graph_string):
    class GeglBuffer(Structure):
        pass
    drawable_id = drawable.ID
    
    gegl = load_library ('libgegl-0.4')
    sucess = gegl.gegl_init (None, None)
    gimp.gimp_drawable_get_shadow_buffer.restype = POINTER (GeglBuffer)
    gimp.gimp_drawable_get_buffer.restype        = POINTER (GeglBuffer)

    x = c_int (); y = c_int (); w = c_int (); h = c_int ()
    non_empty,x,y,w,h = pdb.gimp_drawable_mask_intersect (drawable)
    args = [b"string", c_char_p (gegl_graph_string), c_void_p ()]
    
    if non_empty:
        source = gimp.gimp_drawable_get_buffer (drawable_id)
        target = gimp.gimp_drawable_get_shadow_buffer (drawable_id)
        sucess = gegl.gegl_render_op (source, target, "gegl:gegl", *args)
        gegl.gegl_buffer_flush (target)
        gimp.gimp_drawable_merge_shadow (drawable_id, PushUndo = True)
        gimp.gimp_drawable_update (drawable_id, x, y, w, h)
        gimp.gimp_displays_flush ()
        
# here the single GEGL functions

def gegl_c2g(image, drawable, uradius, usamples, uiter):
    radius=int(uradius)      #dft 300   #ui_range (2, 1000)
    samples=int(usamples)    #dft 4     #ui_range (3, 17)
    iterations=int(uiter)    #dft 10    #ui_range (1, 30)
    enhance_shadows=False
    mystring="gegl:c2g radius=%d samples=%d iterations=%d enhance-shadows=%d"
    gegl_graph_string=mystring % (radius, samples, iterations, enhance_shadows)
    gegl_graph(image, drawable,gegl_graph_string)

def gegl_edge(image, drawable, ualgorithm): 
    algorithm="sobel" # 0=sobel,1=prewitt,2=gradient,3=roberts,4=differential,5=laplace
    amount=2.0        # value_range  (1.00 to 10.0)
    border=1          # 0=none,1=clamp,2=loop,3=black,4=white
    mystring="gegl:edge algorithm=%s amount=%f border=%d"
    gegl_graph_string=mystring % (algorithm, amount, border)
    gegl_graph(image, drawable,gegl_graph_string)

def gegl_median_blur (image, drawable, uradius):
    radius=float(uradius)
    percentile=50.0
    alpha_percentile=50.0
    gegl_graph_string="gegl:median-blur radius=%f percentile=%f alpha-percentile=%f" % (radius, percentile, alpha_percentile)
    gegl_graph(image, drawable, gegl_graph_string)

def gegl_gaussian_blur (image, drawable, ux, uy):    # gaussian blur  
    std_dev_x=float(ux)    #1.50     #user_range   (0.24, 100.0)
    std_dev_y=float(uy)    #1.50     #user_range   (0.24, 100.0)
    mystring="gegl:gaussian-blur std-dev-x=%f std-dev-y=%f"
    gegl_graph_string=mystring % (std_dev_x, std_dev_y)
    gegl_graph(image, drawable,gegl_graph_string)

def gegl_hue_chroma (image, drawable, uhue, uchroma, ulightness):    # Hue-Chroma  
    hue=uhue                #0.0   #value_range  (-180.0, 180.0) 
    chroma=uchroma          #12.0  #value_range  (-100.0, 100.0)  
    lightness=ulightness    #0.0   #value_range  (-100.0, 100.0)
    mystring="gegl:hue-chroma hue=%d chroma=%d lightness=%d"
    gegl_graph_string=mystring % (hue, chroma, lightness)
    gegl_graph(image, drawable, gegl_graph_string)

def gegl_saturation (image, drawable, uscale, ucolorspace):    # saturation 
    scale=uscale            #1.5  #user_range   (0.0, 2.0)
    colorspace=ucolorspace  #0    #value_range  (0, 1, 2)  Native, CIE-lab/lch, CIE-iuv
    mystring="gegl:saturation scale=%f colorspace=%s"
    gegl_graph_string=mystring % (scale, colorspace)
    gegl_graph(image, drawable, gegl_graph_string)

def gegl_exposure (image, drawable, ublack, uexpo):      
    black_level=ublack    # value_range (-0.1, 0.1)
    exposure=uexpo        # ui_range    (-10.0, 10.0)
    gegl_graph_string="gegl:exposure black-level=%f exposure=%f" % (black_level, exposure)
    gegl_graph(image, drawable,gegl_graph_string)
    
def gegl_unsharp_mask (image, drawable, usdev , uscale, uthreshold):      
    std_dev=usdev           #user_range   (0.0, 40.0)
    scale=uscale            #user_range   (0.0, 10.0)
    threshold=uthreshold    #user_range   (0.0, 1.0) 
    mystring="gegl:unsharp-mask std-dev=%f scale=%f threshold=%f"
    gegl_graph_string=mystring % (std_dev, scale, threshold)
    gegl_graph(image, drawable,gegl_graph_string)
    
############################################################################
# include, then use: erro_msg or info_msg
#---------------------------------------------------------------------------
def erro_msg (msg):
    ###
    currenth = pdb.gimp_message_get_handler()
    pdb.gimp_message_set_handler(0) # box
    pdb.gimp_message ("ERROR: "+msg)
    pdb.gimp_message_set_handler(currenth) # current
    ###
def info_msg (msg):
    ###
    currenth = pdb.gimp_message_get_handler()
    pdb.gimp_message_set_handler(2) # error console
    pdb.gimp_message ("INFO: "+msg)
    pdb.gimp_message_set_handler(currenth) # current
    ###

# from Original drawable create the Output image
#---------------------------------------------------------------------------
def createOutimage(inImage, inLayer):
     
    owidth = inImage.width
    oheight = inImage.height
    otype = inImage.base_type
    oname = inLayer.name
    
    #pdb.gimp_selection_none (inImage)
    
   # define/calculate the scale factor for the (new) output image -------------------------
    realWidthHeight = owidth+oheight      # real size of the input image
    optiWidthHeight = FixWH       # size to work on to get best results
    scale = (optiWidthHeight*1.0)/(realWidthHeight*1.0)      # ratio for scaling

    rwidth=int(owidth*scale)
    rheight=int(oheight*scale)
    
   # create a new image of the decided size ------------------------------------------------
   # --- copy the source layer
   # --- scale
    newImage = pdb.gimp_image_new (rwidth, rheight, otype)
    pdb.gimp_selection_none (newImage)
    newBG = pdb.gimp_layer_new_from_drawable(inLayer, newImage)     # background layer
    newBG.mode = LAYER_MODE_NORMAL
    newBG.name = "NewImage: Background Layer from "+oname
    newBG.opacity = 100.0 
    newImage.add_layer(newBG, 0)
    pdb.gimp_layer_scale(newBG, rwidth, rheight, False)
    
   # keep a copy of the original background below the one which will be used ----------------
   # --- and set its visibility off (it should not enter in the process)
    Original = newBG.copy()
    newImage.add_layer(Original, 1)     # below
    pdb.gimp_item_set_visible (Original, False)
    Original.name = oname+" (original)"
    pdb.gimp_layer_add_alpha (newBG)    # RGBA
    # adjust mean values
    adjustmeanv (newImage, newBG)
    # remove low values, remove high values  
    pdb.gimp_drawable_levels (newBG, HISTOGRAM_VALUE, 0.0, 1.0, False,
                                   1.00, 0.20, 0.90, False)
    newBG.opacity = 100.0
    
    return(newImage)

#---------------------------------------------------------------------------
# adjust values
#---------------------------------------------------------------------------
def adjustmeanv(inImage, inLayer):
    # new version modified to take care of the existing mean in determining the theretical mean 
    # check mean value of the layer
    (meanl,std,med,pix,cnt,per) = pdb.gimp_drawable_histogram (inLayer,
                                  HISTOGRAM_VALUE, 0.0, 1.0)
    # consider 3 classes
    if meanl < 85: thmean = (128+meanl)/2  #100
    elif meanl > 170: thmean = (127+meanl)/2  #155
    else: thmean = 127.5    # regular
    minmean = thmean-16
    maxmean = thmean+16
    
    modified=False
    if meanl < minmean and thmean!=meanl:
        ratio = thmean/meanl*1.0 
        #info_msg("mean value found = "+str(meanl)+" ratio = "+str(ratio))
        loop=0
        while meanl < minmean and loop<10:  
            pdb.gimp_drawable_curves_spline (inLayer, HISTOGRAM_VALUE, 8,
                                             (0.0,0.0,0.3,0.3*ratio, 0.7,0.7*ratio, 1.0,1.0))
            (meanl,std,med,pix,cnt,per) = pdb.gimp_drawable_histogram (inLayer,
                                  HISTOGRAM_VALUE, 0.00, 1.0)
            loop+=1
        modified=True
        #info_msg("increased mean value = "+str(meanl))
    if meanl > maxmean and thmean!=meanl:
        ratio = thmean/meanl*1.0 
        #info_msg("mean value found = "+str(meanl))
        loop=0
        while meanl > maxmean and loop<10:  
            pdb.gimp_drawable_curves_spline (inLayer, HISTOGRAM_VALUE, 8,
                                             (0.0,0.0,0.3,0.3*ratio, 0.7,0.7*ratio, 1.0,1.0))
            (meanl,std,med,pix,cnt,per) = pdb.gimp_drawable_histogram (inLayer,
                                  HISTOGRAM_VALUE, 0.00, 1.0)
            loop+=1
        mofified=True
        #info_msg("reduced mean value = "+str(meanl))
    
    return(inLayer)
     
#
# ================ *
#   M  A  I  N     *
# ---------------- *
   
def JvidEffectMG (inImage, inLayer, widthheight, texturize,
                  top, masked, bwcol, pluscol, strokes,
                  doFlatten):
    
    # house-keeping
    oname = inLayer.name
    if "." in oname:
        split = oname.split(".", 1)
        oname = split[0]
        
    OrigWH = inLayer.width+inLayer.height
    
    ts = time.time()
    st = datetime.datetime.fromtimestamp(ts).strftime("%Y-%m-%d %H-%M-%S")
    
    # parms used
    parmsused = str(texturize)+","+str(top)+","+str(masked)+","+str(bwcol)+","+str(pluscol)+","+str(strokes)
    info_msg ("parms used = "+parmsused)
    
    newImage = createOutimage (inImage, inLayer)
    Original = newImage.layers[1]
    Original.name = "Original"
    Preprocessed = newImage.layers[0]
    Preprocessed.name = "Preprocessed"
    width = Preprocessed.width
    height = Preprocessed.height
    # warning: G'MIC DOODLE does not handle correctly a selection !!!
    # hence we remove it if any
    # take care of a possible existing selection
    NOsel = pdb.gimp_selection_is_empty(newImage)
    if NOsel == False:
        pdb.gimp_selection_all (newImage)
    
    #pdb.gimp_display_new (newImage)
    #return
    
    # prepare image for effects Blur + USM 
    # - Blur  
    BlurSharp = Preprocessed.copy()
    newImage.add_layer (BlurSharp, 0)
    uRadius=3.0
    BlurSharp.name = "Median Blurred 6.0" 
    gegl_median_blur (newImage, BlurSharp, uRadius)
    # - USM
    uSdev=6.0; uScale=2.0; uThreshold=0.0       # new parameters
    BlurSharp.name = "Median Blurred 3.0 then USM 6.0/2.0" 
    gegl_unsharp_mask (newImage, BlurSharp, uSdev, uScale, uThreshold)
    
    InputDoodle = BlurSharp.copy()
    newImage.add_layer (InputDoodle, 0)
    InputDoodle.opacity = 100.0
    if texturize==0:    # no texturize 
        InputDoodle.name = "Input layer for Doodle not texturized"
        variant="no texturize"
    if texturize==1:    # texturize before doodle
        info_msg ("applying Barbouillage before, please wait...)")
        variant="barbouillage before"
        pdb.plug_in_gmic_qt(newImage, InputDoodle, 1,0,
            "-v - -samj_Barbouillage_Paint_Daub 2,1,100,0.1,1,4,1,0,8  ")
        InputDoodle.name = "Input layer for Doodle texturized" 
    #[G'MIC] Doodle: fx_doodle 30,2,2,1.5,2,70,1 default
    info_msg ("applying Doodle, please wait...)")
    # warning: Doodle is slow, hence reduce temporatily the size
    pdb.gimp_image_scale (newImage, width/2, height/2)    
    pdb.plug_in_gmic_qt(newImage, InputDoodle, 1,2,
    ####"-v - -fx_doodle 30,2,2,1.5,1,35,0 ")   # no preview
        "-v - -fx_doodle 20,2,2,1.0,1,35,0 ")   # no preview    #        try
    Doodle = newImage.layers[0]
    Doodle.opacity = 100.0
    Doodle.name = "Doodle Layer -fx_doodle 30,2,2,1.5,1,35,0 "
    Doodle.mode = LAYER_MODE_HSV_VALUE
    pdb.gimp_image_scale (newImage, width, height) # restore
    if strokes==0:
        SPREAD=FixWH*3.0/1000.0
    if strokes==1:
        SPREAD=FixWH*2.0/1000.0
    if strokes==2:
        SPREAD=FixWH*1.0/1000.0
    pdb.plug_in_spread (newImage, Doodle, SPREAD, SPREAD)
    # check whether user like to texturize image after doodling
    if texturize==2:  # texturize after doodle
        info_msg ("applying Barbouillage after, please wait...)")
        TextureAfter = BlurSharp.copy()
        newImage.add_layer (TextureAfter, 0)
        TextureAfter.name = "Texture layer after Doodle"
        TextureAfter.mode = LAYER_MODE_NORMAL
        TextureAfter.opacity = 50.0
        variant="barbouillage after"
        pdb.plug_in_gmic_qt(newImage, TextureAfter, 1,0,
            "-v - -samj_Barbouillage_Paint_Daub 2,1,100,0.1,1,4,1,0,8  ")
    
    #pdb.gimp_display_new (newImage) 
    #return
    
    # prepare outcome  
    Base1 = pdb.gimp_layer_new_from_visible (newImage, newImage, "")
    newImage.add_layer (Base1, 0)
    Base1.name = "Base for Desaturate (C2G 500,5,10)"
    Base1.mode = LAYER_MODE_NORMAL
    #pdb.gimp_drawable_desaturate (Base1, DESATURATE_VALUE)
    RADIUS=500; SAMPLES=5; ITER=10
    gegl_c2g (newImage, Base1, RADIUS, SAMPLES, ITER)
    
    #pdb.gimp_display_new (newImage)
    #return
    
    # top
    if top==0:
        OnTop = Original.copy()
        newImage.add_layer (OnTop, 0)
        pdb.gimp_item_set_visible (OnTop, True)
        OnTop.name = "Original on top"
    if top==1:
        OnTop = Preprocessed.copy()
        newImage.add_layer (OnTop, 0)
        OnTop.name = "Preprocessed on top"
    if top==2:
        OnTop = Original.copy()
        newImage.add_layer (OnTop, 0)
        pdb.gimp_item_set_visible (OnTop, True)
        pdb.gimp_drawable_posterize (OnTop, 8)
        OnTop.name = "Posterized on top"
        
    if pluscol==0:      
        SCALE=1.75; COLORSPACE="native"
        gegl_saturation (newImage, OnTop, SCALE, COLORSPACE)
    if pluscol==1:      
        UHUE=0.0; UCHROMA=20.0; ULIGHTNESS=0.0
        gegl_hue_chroma (newImage, OnTop, UHUE, UCHROMA, ULIGHTNESS)
    if pluscol==2:      
        SCALE=1.25; COLORSPACE="native"
        gegl_saturation (newImage, OnTop, SCALE, COLORSPACE)
        UHUE=0.0; UCHROMA=12.0; ULIGHTNESS=0.0
        gegl_hue_chroma (newImage, OnTop, UHUE, UCHROMA, ULIGHTNESS) 
        
    if masked==0: 
        mask = pdb.gimp_layer_create_mask (OnTop, ADD_MASK_WHITE)
        pdb.gimp_layer_add_mask (OnTop, mask)
    if masked==1:
        mask = pdb.gimp_layer_create_mask (OnTop, ADD_MASK_COPY)
        pdb.gimp_layer_add_mask (OnTop, mask)
        pdb.gimp_drawable_curves_spline (mask, HISTOGRAM_VALUE, 4, (0.0,0.75,1.0,1.0))
    if masked==2:
        mask = pdb.gimp_layer_create_mask (OnTop, ADD_MASK_COPY)
        pdb.gimp_layer_add_mask (OnTop, mask)
        pdb.gimp_drawable_invert (mask, TRUE)
        pdb.gimp_drawable_curves_spline (mask, HISTOGRAM_VALUE, 4, (0.0,0.75,1.0,1.0))
        
    if bwcol==0:    # grayscale
        OnTop.mode = LAYER_MODE_VIVID_LIGHT
        pdb.gimp_drawable_desaturate (OnTop, DESATURATE_LUMA)
    if bwcol==1:    # light colours
        OnTop.mode = LAYER_MODE_HARDLIGHT
    if bwcol==2:    # vivid colours
        OnTop.mode = LAYER_MODE_HARDLIGHT
    
    #pdb.gimp_display_new (newImage)
    #return
    
    # final outcome  
    Base2 = pdb.gimp_layer_new_from_visible (newImage, newImage, "")
    newImage.add_layer (Base2, 0)
    Base2.name = "Preliminary Outcome for multiply "
    if bwcol==0:
        Base2.mode = LAYER_MODE_LINEAR_BURN
        Base2.opacity = 55.0
    if bwcol==1:
        Base2.mode = LAYER_MODE_SCREEN
        Base2.opacity = 25.0
    if bwcol==2:
        Base2.mode = LAYER_MODE_MULTIPLY
        Base2.opacity = 99.0
        
    #pdb.gimp_display_new (newImage)
    #return
    
    if doFlatten==True:
        outcome = pdb.gimp_image_merge_visible_layers (newImage, CLIP_TO_BOTTOM_LAYER)
        outcome.name = "outcome"
    else:    
        outcome = pdb.gimp_layer_new_from_visible (newImage, newImage, "Outcome")
        newImage.add_layer (outcome, 0)
        pdb.gimp_item_set_visible (outcome, False)
    outcome.name = st+" "+oname+" with JVID effect MG (Doodle Draw Graphic Effect) v.G "+parmsused
    
    if widthheight==0: widthheight=OrigWH
    uRatio = widthheight*1.00/FixWH*1.00
    uWidth = int(width*uRatio)
    uHeight = int(height*uRatio)
    #info_msg ("SCALING")   # not during tests, scaling is slow
    pdb.gimp_image_scale (newImage, uWidth, uHeight)   # to user chosen size
    
    pdb.gimp_display_new (newImage)

### Registration
whoiam="\n"+os.path.abspath(sys.argv[0])
author="issabella/dinasset"
menu="<Image>/Jvid/"
copyrightYear="2021"
desc="JVID Effect MG (Doodle Draw Graphic Effect) v.G "

register(
    "JvidEffectMG",
    desc+whoiam,desc,
    author,author,
    copyrightYear,
    desc+" ...",
    "*",
    [
        (PF_IMAGE, "image", "Input image", None),
        (PF_DRAWABLE, "drawable", "Input drawable", None),
        (PF_ADJUSTMENT, "newsize", "Width+Height of New Image"+"\n (0=same as input)",
                                    3000, (0, 6000, 100)),
        (PF_OPTION, "texturize", "Texturize", 2, ["*none",
                                                  "Barbouillage before",
                                                  "Barbouillage after"]),
        (PF_OPTION, "top", "Top layer", 1, ["original",
                                            "processed",
                                            "poster"]),
        (PF_OPTION, "mask", "Top mask", 0, ["white",
                                            "gray-copy",
                                            "gray-copy inverted"]),
        (PF_OPTION, "bwcol", "BW/Colours", 2, ["grayscale",
                                               "light colours",
                                               "vivid colours"]),
        (PF_OPTION, "pluscol", "Plus Colour", 2, ["add saturation",
                                                  "add chroma",
                                                  "both"]),
        (PF_OPTION, "strokes", "Strokes", 1, ["soft",
                                              "normal",
                                              "hard"]),
        (PF_TOGGLE, "flatten", "Flatten Image?", False),
    ],
    [],
    JvidEffectMG,
    menu=menu
)

main()
